/***************************************************************************

    file                 : spline.h
    created              : Wed Mai 14 19:53:00 CET 2003
    copyright            : (C) 2003-2004 by Bernhard Wymann
    email                : berniw@bluewin.ch
    version              : $Id: spline.h,v 1.3 2004/08/10 20:11:49 berniw Exp $

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _RACELINE_H_
#define _RACELINE_H_

#include <iostream>
#include <fstream>
#include <string>

#include "linalg.h"

enum { LINE_MID=0, LINE_RL, LINE_CLOTH_PRIV };

#define NUM_RACELINES 3
#define MAXSEGMENTS 4000
#define MAXDIVS 6000

class LRaceLine {
 public:
  LRaceLine()
  {
   fDirt = 0;
  }
  ~LRaceLine();

  void setCornerSpeed( double wi ) { CornerSpeed = wi; }
  void setBrakeDist( double wi ) { BrakeDist = wi; }
  //void setbrakeDist( double wi ) { brakeDist = wi; }
  //void setspeedAdjust( double wi ) { speedAdjust = wi; }
  //void setminTurnInverse( double wi ) { minTurnInverse = wi; }
  void setCar( tCarElt *mycar ) { car = mycar; }
  void setRwData( tTrack* t, void **carParmHandle, tSituation *s);

  double minTurnInverse;
  double CornerSpeed;
  double turnSpeed;
  double brakeDist;
  double BrakeDist;
  double IntMargin;
  double ExtMargin;
  double speedAdjust;
  double wheelbase;
  double wheeltrack;

  int Divs;
  int DivLength;
  int Segs;

  double Width;
  double Length;
  double TargetSpeed;
  double tSegDist[MAXSEGMENTS];
  int tSegIndex[MAXSEGMENTS];
  double tElemLength[MAXSEGMENTS];
  double tx[NUM_RACELINES][MAXDIVS];
  double ty[NUM_RACELINES][MAXDIVS];
  double tLane[NUM_RACELINES][MAXDIVS];
  double tRInverse[NUM_RACELINES][MAXDIVS];
  double tSpeed[NUM_RACELINES][MAXDIVS];
  double tDistance[MAXDIVS];
  double tMaxSpeed[MAXDIVS];
  double txLeft[MAXDIVS];
  double tyLeft[MAXDIVS];
  double txRight[MAXDIVS];
  double tyRight[MAXDIVS];
  double tFriction[MAXDIVS];
  double tzLeft[MAXDIVS];
  double tzRight[MAXDIVS];

  int fDirt;
  int Next;
  int This;

  tCarElt *car;

  void UpdateTxTy(int i, int rl);
  void SetSegmentInfo(const tTrackSeg *pseg, double d, int i, double l);
  void SplitTrack(tTrack *ptrack, int rl);
  double GetRInverse(int prev, double x, double y, int next, int rl);
  double getRInverse() { return tRInverse[LINE_RL][Next]; }
  double getRInverse(double distance) { int d = ((Next + int(distance/DivLength)) % Divs); return tRInverse[LINE_RL][d]; }
  void AdjustRadius(int prev, int i, int next, double TargetRInverse, int rl, double Security = 0);
  void Smooth(int Step, int rl);
  void StepInterpolate(int iMin, int iMax, int Step, int rl);
  void Interpolate(int Step, int rl);
  void InitTrack(tTrack* track, tSituation *p);
  void NewRace(tCarElt* newcar, tSituation *s);
  void GetRaceLineData(tSituation *s, v2d *target, float *speed, float *avspeed, float *raceoffset, float *lookahead, float *racesteer);
  void GetPoint( float offset, float lookahead, vec2f *rt );
  int isOnLine( int line);
  double correctLimit(int line);
  double getAvoidSpeed( float distance1, float distance2 );
  double getCornerSpeed( int Div );
  double getBrakeDist( int Div );
  double getIntMargin( int Div );
  double getExtMargin( int Div );

#define MAX_TDATA 10
  	int turnSpeedSegStart[MAX_TDATA];
  	int turnSpeedSegEnd[MAX_TDATA];
  	float segTurnSpeed[MAX_TDATA];

  	int brakeDistSegStart[MAX_TDATA];
  	int brakeDistSegEnd[MAX_TDATA];
  	float segBrakeDist[MAX_TDATA];

  	int MspeedSegStart[MAX_TDATA];
  	int MspeedSegEnd[MAX_TDATA];
  	float segMaxSpeed[MAX_TDATA];

  	int extMargSegStart[MAX_TDATA];
  	int extMargSegEnd[MAX_TDATA];
  	float segExtMargin[MAX_TDATA];

  	int intMargSegStart[MAX_TDATA];
  	int intMargSegEnd[MAX_TDATA];
  	float segIntMargin[MAX_TDATA];

  int line_verbose;
  int steer_verbose;
  int LineIndex;
  int LineSave;
  int rlLine;

  private:
  // Utility functions
  	void saveFile();
	void StoreData(tTrack* track);
	void WriteLine(tTrack* track);
	void WriteTrack(tTrack* track);
};

#endif // _RACELINE_H_

